#include <u.h>
#include <libc.h>

#include "vga.h"

void
dbdumpmode(Mode* mode)
{
	Attr *attr;

	fprintf(stdout, "dbdumpmode\n");

	fprintf(stdout, "type=%s, size=%s\n", mode->type, mode->size);
	fprintf(stdout, "frequency=%d\n", mode->frequency);
	fprintf(stdout, "x=%d (0x%X), y=%d (0x%X), z=%d (0x%X)\n",
		mode->x, mode->x, mode->y,  mode->y, mode->z, mode->z);
	fprintf(stdout, "ht=%d (0x%X), shb=%d (0x%X), ehb=%d (0x%X)\n",
		mode->ht, mode->ht, mode->shb, mode->shb, mode->ehb, mode->ehb);
	fprintf(stdout, "shs=%d (0x%X), ehs=%d (0x%X)\n",
		mode->shs, mode->shs, mode->ehs, mode->ehs);
	fprintf(stdout, "vt=%d (0x%X), vrs=%d (0x%X), vre=%d (0x%X)\n",
		mode->vt, mode->vt, mode->vrs, mode->vrs, mode->vre, mode->vre);
	fprintf(stdout, "hsync=%d, vsync=%d, interlace=%d\n",
		mode->hsync, mode->vsync, mode->interlace);

	for(attr = mode->attr; attr; attr = attr->next)
		fprintf(stdout, "mode->attr: %s=%s\n", attr->attr, attr->val);
}

char*
dbattr(Attr *a, char *s)
{
	return nil;
}

static Ctlr*
addctlr(Vga* vga, char* val)
{
	Ctlr **ctlr;
	char name[NAMELEN+1], *p;
	int i;

	if(val == nil)
		return nil;

	/*
	 * A controller name may have an extension on the end
	 * following a '-' which can be used as a speed grade or
	 * subtype.  Do the match without the extension.
	 * The linked copy of the controller struct gets the
	 * full name with extension.
	 */
	strncpy(name, val, NAMELEN);
	name[NAMELEN] = 0;
	if(p = strchr(name, '-'))
		*p = 0;

	for(i = 0; ctlrs[i]; i++){
		if(strcmp(ctlrs[i]->name, name))
			continue;
		for(ctlr = &vga->link; *ctlr; ctlr = &((*ctlr)->link))
			;
		*ctlr = alloc(sizeof(Ctlr));
		**ctlr = *ctlrs[i];
		strncpy((*ctlr)->name, val, NAMELEN);
		return *ctlr;	
	}

	error("dbctlr: unknown controller \"%s\" ctlr\n", val);
	return 0;
}

long
dbctlr(char *name, Vga *vga)
{
	vga->ctlr = addctlr(vga, name);
	vga->ramdac = addctlr(vga, ramdacstr);
	vga->clock = addctlr(vga, clockstr);
	vga->hwgc = addctlr(vga, hwgcstr);
	addctlr(vga, linkstr);
	return 0;
}
