#include <u.h>
#include <libc.h>

#include "vga.h"
#include "pci.h"

#define BY2PG 4096

/*
 * ATI Mach64(CT|ET|G*|VT|VU).
 */
enum {
	HTotalDisp,
	HSyncStrtWid,
	VTotalDisp,
	VSyncStrtWid,
	VlineCrntVline,
	OffPitch,
	IntCntl,
	CrtcGenCntl,
	LcdGenCntl,		/* LT */
	PowerMngmnt,	/* LT */

	OvrClr,
	OvrWidLR,
	OvrWidTB,

	CurClr0,
	CurClr1,
	CurOffset,
	CurHVposn,
	CurHVoff,

	ScratchReg0,
	ScratchReg1,	/* Scratch Register (BIOS info) */
	ClockCntl,
	BusCntl,
	MemCntl,
	MemVgaWpSel,
	MemVgaRpSel,
	DacRegs,
	DacCntl,
	GenTestCntl,
	ConfigCntl,	/* Configuration control */
	ConfigChipId,
	ConfigStat0,	/* Configuration status 0 */
	ConfigStat1,	/* Configuration status 1 */
	DspConfig,	/* Rage */
	DspOnOff,	/* Rage */

	DpBkgdClr,
	DpChainMsk,
	DpFrgdClr,
	DpMix,
	DpPixWidth,
	DpSrc,
	DpWriteMsk,
	Nreg,

	LcdIndex = 0x29,	/* LTPro */
	LcdData = 0x2A,
	TvIndex = 0x1D,
	TvData = 0x27,

	LCD_ConfigPanel = 0,
	LCD_GenCntl,
	LCD_DstnCntl,
	LCD_HfbPitchAddr,
	LCD_HorzStretch,
	LCD_VertStretch,
	LCD_ExtVertStretch,
	LCD_LtGio,
	LCD_PowerMngmnt,
	LCD_ZvgPio,
	Nlcd,
};


static ushort ioregs[Nreg] = {
 0x0000, 0x0100, 0x0200, 0x0300, 0x0400, 0x0500, 0x0600, 0x0700,
 0x0800, 0x0900, 0x0a00, 0x0b00, 0x0c00, 0x0d00, 0x0e00, 0x0f00,
 0x1000, 0x1100, 0x1200, 0x1300, 0x1400, 0x1500, 0x1600, 0x1700,
 0x1800, 0x1900, 0x1a00, 0x1b00, 0x1c00, 0x1d00, 0x1e00, 0x1f00,
};

static ushort pciregs[Nreg] = {
  /* HTotalDisp */		0x00,
  /* HSyncStrtWid */	0x01,
  /* VTotalDisp */		0x02,
  /* VSyncStrtWid */        0x03,
  /* VlineCrntVline */      0x04,
  /* OffPitch */		0x05,
  /* IntCntl */			0x06,
  /* CrtcGenCntl */		0x07,
  /* LcdGenCntl */		0x35,
  /* PowerMngmnt */		0x36,

  /* OvrClr */			0x10,
  /* OvrWidLR */		0x11,
  /* OvrWidTB */		0x12,

  /* CurClr0 */		0x18,
  /* CurClr1 */		0x19,
  /* CurOffset */		0x1A,
  /* CurHVposn */		0x1B,
  /* CurHVoff */		0x1C,

  /* ScratchReg0 */		0x20,
  /* ScratchReg1 */		0x21,
  /* ClockCntl */		0x24,
  /* BusCntl */			0x28,
  /* MemCntl */		0x2C,
  /* MemVgaWpSel */	0x2D,
  /* MemVgaRpSel */	0x2E,
  /* DacRegs */		0x30,
  /* DacCntl */		0x31,
  /* GenTestCntl */		0x34,
  /* ConfigCntl */		0x37,
  /* ConfigChipId */	0x38,
  /* ConfigStat0 */		0x39,
  /* ConfigStat1 */		0x3A,

  /* DspConfig */		0x08,
  /* DspOnOff */		0x09,

  /* DpBkgdClr */		0xB0,
  /* DpChainMsk */		0xB3,
  /* DpFrgdClr */		0xB1,
  /* DpMix */			0xB5,
  /* DpPixWidth */		0xB4,
  /* DpSrc */			0xB6,
  /* DpWriteMsk */		0xB2,
};

enum {
	PLLm		= 0x02,
	PLLp		= 0x06,
	PLLn2		= 0x09,
	PLLx            = 0x0b,         /* external divisor (Rage) */

	Npll		= 32,
};

typedef struct Mach64xx	Mach64xx;
struct Mach64xx {
	ulong	io;
	ulong	mmregbase;
	ushort* regtable;
	Pcidev*	pci;
	long		bigmem;

	ulong	reg[Nreg];
	uchar	pll[Npll];
	ulong	lcd[Nlcd];

	ulong	(*ior32)(Mach64xx*, long);
	void		(*iow32)(Mach64xx*, long, ulong);
};

static uchar
pllr(Mach64xx* mp, long r)
{
	long io;

	io = (mp->regtable[ClockCntl]<<2)+mp->io;
	outportb(io+1, r<<2);
	return inportb(io+2);
}

static void
pllw(Mach64xx* mp, long r, uchar b)
{
	long io;

	io = (mp->regtable[ClockCntl]<<2)+mp->io;
	outportb(io+1, (r<<2)|0x02);
	outportb(io+2, b);
}

static ulong
portior32(Mach64xx* mp, long r)
{
	return inportl((mp->regtable[r]<<2)+mp->io);
}

static void
portiow32(Mach64xx* mp, long r, ulong l)
{
	outportl((mp->regtable[r]<<2)+mp->io, l);
}

static ulong
mmior32(Mach64xx* mp, long r)
{
	return *(ulong*)(mp->mmregbase+(mp->regtable[r]<<2));
}

static void
mmiow32(Mach64xx* mp, long r, ulong l)
{
	*(ulong*)(mp->mmregbase+(mp->regtable[r]<<2)) = l;
}

static ulong
lcdr32(Mach64xx *mp, ulong r)
{
	outportb(mp->io+(LcdIndex<<2), r&0x0F);
	return inportl(mp->io+(LcdData<<2));
}

static void
lcdw32(Mach64xx *mp, ulong r, ulong v)
{
	outb(mp->io+(LcdIndex<<2), r&0x0F);
	outportl(mp->io+(LcdData<<2), v);
}

static ulong
tvr32(Mach64xx *mp, ulong r)
{
	outportb(mp->io+(TvIndex<<2), r&0x0F);
	return inportl(mp->io+(TvData<<2));
}

static void
tvw32(Mach64xx *mp, ulong r, ulong v)
{
	outb(mp->io+(TvIndex<<2), r&0x0F);
	outportl(mp->io+(TvData<<2), v);
}

long smallmem[] = {
	512*1024, 1024*1024, 2*1024*1024, 4*1024*1024,
	6*1024*1024, 8*1024*1024, 12*1024*1024, 16*1024*1024,
};

long bigmem[] = {
	512*1024, 2*512*1024, 3*512*1024, 4*512*1024,
	5*512*1024, 6*512*1024, 7*512*1024, 8*512*1024,
	5*1024*1024, 6*1024*1024, 7*1024*1024, 8*1024*1024,
	10*1024*1024, 12*1024*1024, 14*1024*1024, 16*1024*1024
};

static void
snarf(Vga* vga, Ctlr* ctlr)
{
	Mach64xx *mp;
	long i, m;
	ulong v;

	if(vga->private == nil){
		vga->private = alloc(sizeof(Mach64xx));
		mp = vga->private;
		mp->io = 0x2EC;
		mp->ior32 = portior32;
		mp->iow32 = portiow32;
		mp->regtable = ioregs;
		mp->pci = pcimatch(0, 0x1002, 0);
		if (mp->pci) {
			if(v = mp->pci->mem[1].bar & ~0x3) {
				mp->io = v;
				mp->regtable = pciregs;
			}
/*			if((v = segattach(0, "mach64mmio", nil, 0x400)) != ~0) {
				print("have pci mem %lux\n", v);
				mp->mmregbase = v+BY2PG-0x400;
				mp->ior32 = mmior32;
				mp->iow32 = mmiow32;
				mp->regtable = pciregs;
			}
*/
		}
	}

	mp = vga->private;
	for(i = 0; i < Nreg; i++)
		mp->reg[i] = mp->ior32(mp, i);

	for(i = 0; i < Npll; i++)
		mp->pll[i] = pllr(mp, i);

	for(i = 0; i < Nlcd; i++)
		mp->lcd[i] = lcdr32(mp, i);

	/*
	 * Check which memory size map we are using.
	 */
	mp->bigmem = 0;
	switch (mp->reg[ConfigChipId]&0xffff) {
		case ('G'<<8)|'B':	/* 4742: 264GT PRO */
		case ('G'<<8)|'D':	/* 4744: 264GT PRO */
		case ('G'<<8)|'I':		/* 4749: 264GT PRO */
		case ('G'<<8)|'P':	/* 4750: 264GT PRO */
		case ('G'<<8)|'Q':	/* 4751: 264GT PRO */
		case ('G'<<8)|'U':	/* 4755: 264GT DVD */
		case ('V'<<8)|'U':	/* 5655: 264VT3 */
		case ('L'<<8)|'P':	/* 4C50: Rage LTPro */
			mp->bigmem = 1;
		case ('G'<<8)|'T':	/* 4754: 264GT[B] */
		case ('V'<<8)|'T':	/* 5654: 264VT/GT/VTB */
			/*
			 * only the VTB and GTB use the new memory encoding,
			 * and they are identified by a nonzero ChipVersion, apparently.
			 */
			if((mp->reg[ConfigChipId] >> 24) & 0x7)
				mp->bigmem = 1;
			break;
	}

	/*
	 * Memory size and aperture. It's recommended
	 * to use an 8Mb aperture on a 16Mb boundary.
	 */

	if(mp->bigmem)
		vga->vmz = bigmem[mp->reg[MemCntl] & 0x0F];
	else
		vga->vmz = smallmem[mp->reg[MemCntl] & 0x07];
	vga->vma = 16*1024*1024;

	m = mp->pll[PLLm];
	vga->f[1] = (2*255*(double)RefFreq)/m;

	ctlr->flag |= Fsnarf;
}

static void
options(Vga *vganotused, Ctlr* ctlr)
{
	ctlr->flag |= Hlinear|Foptions;
}

static void
clock(Vga* vga, Ctlr* ctlr)
{
	long m, p;
	double f, q;
	Mach64xx *mp;

	mp = vga->private;

	if(vga->f[0] == 0)
		vga->f[0] = vga->mode->frequency;
	f = vga->f[0];

	/*
	 * To generate a specific output frequency, the reference (m),
	 * feedback (n), and post dividers (p) must be loaded with the
	 * appropriate divide-down ratios. In the following r is the
	 * XTALIN frequency (usually RefFreq) and t is the target frequency
	 * (vga->f).
	 *
	 * Use the maximum reference divider left by the BIOS for now,
	 * otherwise MCLK might be a concern. It can be calculated as
	 * follows:
	 * 			    Upper Limit of PLL Lock Range
	 *	Minimum PLLREFCLK = -----------------------------
	 *				      (2*255)
	 *
	 *				       XTALIN
	 *			m = Floor[-----------------]
	 *				  Minimum PLLREFCLK
	 *
	 * For an upper limit of 135MHz and XTALIN of 14.318MHz m
	 * would be 54.
	 */
	m = mp->pll[PLLm];
	vga->m[0] = m;

	/*
	 * The post divider may be 1, 2, 4 or 8 and is determined by
	 * calculating
	 *		 t*m
	 *	    q = -----
	 *		(2*r)
	 * and using the result to look-up p.
	 */
	q = (f*m)/(2*RefFreq);
	if(ctlr->flag&Uenhanced){
	  if(q > 255 || q < 10.6666666667)
		error("%s: vclk %d out of range\n", ctlr->name, vga->f[0]);
	  if(q > 127.5)
		p = 1;
	  else if(q > 85)
		p = 2;
	  else if(q > 63.75)
		p = 3;
	  else if(q > 42.5)
		p = 4;
	  else if(q > 31.875)
		p = 6;
	  else if(q > 21.25)
		p = 8;
	  else
		p = 12;
	}else{
	  if(q > 255 || q < 16)
		error("%s: vclk %d out of range\n", ctlr->name, vga->f[0]);
	  if(q >= 127.5)
		p = 1;
	  else if(q >= 63.5)
		p = 2;
	  else if(q >= 31.5)
		p = 4;
	  else
		p = 8;
	}
	vga->p[0] = p;

	/*
	 * The feedback divider should be kept in the range 0x80 to 0xFF
	 * and is found from
	 *	n = q*p
	 * rounded to the nearest whole number.
	 */
	vga->n[0] = (q*p)+0.5;
}

static void
setdsp(Vga* vga, Ctlr *ctlrnotused)
{
	Mach64xx *mp;
	ushort table, memclk, memtyp;
	long t, page;
	float x, fifoon, fifooff;
	ushort y, loop, dspon, dspoff, xclks;

	mp = vga->private;

	/*
	 * Get video ram configuration from BIOS and chip
	 */
	readbios((char*)&table, sizeof table, 0xc0048);
	readbios((char*)&table, sizeof table, 0xc0000+table+16);
	readbios((char*)&memclk, sizeof memclk, 0xc0000+table+18);
	memtyp = mp->reg[ConfigStat0]&07;
	/* /n/emelieother/jmk/ftp/xfree86.org/xc/programs/Xserver/hw/xfree86/vga256/drivers/ati/atidsp.c:175 */
	/* BUG: this is probably wrong now that we allow more z values.  -rsc */
	x = ((float)memclk*640000.0) / ((float)vga->mode->frequency*(float)vga->mode->z);
	y = (ushort)(24.0*x) + 1;
	for (t = -5; y; t++)
		y >>= 1;

	xclks = x * (1 << (11-t));
	if (vga->vmz > 1024*1024) {
		if (memtyp >= 4) {
			loop = 8;
			page = 8;
		} else {
			loop = 6;
			page = 9;
		}
	} else {
		if (memtyp >= 4) {
			loop = 9;
			page = 10;
		} else {
			loop = 8;
			page = 10;
		}
	}
	mp->reg[DspConfig] = ((t&0x7) << 20) | (loop << 16) | (xclks & 0x3fff);

	if (x >= page)
		fifoon = 2*page + 1 + (long)x;
	else
		fifoon = 3*page;
	fifooff = (24-1)*x + 1;
	dspon  = (ushort)fifoon  << (6-t);
	dspoff = (ushort)fifooff << (6-t);
	mp->reg[DspOnOff] = ((dspon&0x7ff) << 16) | (dspoff&0x7ff);

}

static void
init(Vga* vga, Ctlr* ctlr)
{
	Mode *mode;
	Mach64xx *mp;
	long p;

	mode = vga->mode;
	if((mode->x > 640 || mode->y > 480) && mode->z == 1)
		error("%s: no support for 1-bit mode other than 640x480x1\n", ctlr->name);

	mp = vga->private;
	if(mode->z > 8 && mp->pci == nil)
		error("%s: no support for >8-bit color without PCI\n", ctlr->name);

	/*
	 * Check for Rage chip
	 */
	switch (mp->reg[ConfigChipId]&0xffff) {
		case ('G'<<8)|'B':	/* 4742: 264GT PRO */
		case ('G'<<8)|'D':	/* 4744: 264GT PRO */
		case ('G'<<8)|'I':		/* 4749: 264GT PRO */
		case ('G'<<8)|'P':	/* 4750: 264GT PRO */
		case ('G'<<8)|'Q':	/* 4751: 264GT PRO */
		case ('G'<<8)|'U':	/* 4755: 264GT DVD */
		case ('V'<<8)|'U':	/* 5655: 264VT3 */
		case ('G'<<8)|'T':	/* 4754: 264GT[B] */
		case ('V'<<8)|'T':	/* 5654: 264VT/GT/VTB */
		case ('L'<<8)|'P':	/* 4C50: Rage LTPro */
			ctlr->flag |= Uenhanced;
			break;
	}

	/*
	 * Always use VCLK2.
	 */
	clock(vga, ctlr);
	mp->pll[PLLn2] = vga->n[0];
	mp->pll[PLLp] &= ~(0x03<<(2*2));
	switch(vga->p[0]){
	case 1:
	case 3:
		p = 0;
		break;

	case 2:
		p = 1;
		break;

	case 4:
	case 6:
		p = 2;
		break;

	case 8:
	case 12:
		p = 3;
		break;

	default:
		p = 3;
		break;
	}
	mp->pll[PLLp] |= p<<(2*2);
	if ((1<<p) != vga->p[0])
		mp->pll[PLLx] |= 1<<(4+2);
	else
		mp->pll[PLLx] &= ~(1<<(4+2));
	mp->reg[ClockCntl] = 2;

	mp->reg[ConfigCntl] = 0;

	mp->reg[CrtcGenCntl] = 0x02000000|(mp->reg[CrtcGenCntl] & ~0x01000700);
	switch(mode->z){
	default:
	case 1:
		mp->reg[CrtcGenCntl] |= 0x00000100;
		mp->reg[DpPixWidth] = 0x00000000;
		break;
	case 8:
		mp->reg[CrtcGenCntl] |= 0x01000200;
		mp->reg[DpPixWidth] = 0x00020202;
		break;
	case 15:
		mp->reg[CrtcGenCntl] |= 0x01000300;
		mp->reg[DpPixWidth] = 0x00030303;
		break;
	case 16:
		mp->reg[CrtcGenCntl] |= 0x01000400;
		mp->reg[DpPixWidth] = 0x00040404;
		break;
	case 24:
		mp->reg[CrtcGenCntl] |= 0x01000500;
		mp->reg[DpPixWidth] = 0x00050505;
		break;
	case 32:
		mp->reg[CrtcGenCntl] |= 0x01000600;
		mp->reg[DpPixWidth] = 0x00060606;
		break;
	}

	print("PixWidth %lx\n", mp->reg[DpPixWidth]);

	mp->reg[HTotalDisp] = (((mode->x>>3)-1)<<16)|((mode->ht>>3)-1);
	mp->reg[HSyncStrtWid] = (((mode->ehb - mode->shs)>>3)<<16)|((mode->shs>>3)-1);
	if(mode->hsync == '-')
		mp->reg[HSyncStrtWid] |= 0x00200000;
	mp->reg[VTotalDisp] = ((mode->y-1)<<16)|(mode->vt-1);
	mp->reg[VSyncStrtWid] = ((mode->vre - mode->vrs)<<16)|(mode->vrs-1);
	if(mode->vsync == '-')
		mp->reg[VSyncStrtWid] |= 0x00200000;
	mp->reg[IntCntl] = 0;

	/*
	 * This used to set it to (mode->x/(8*2))<<22 for depths < 8,
	 * but from the manual that seems wrong to me.  -rsc
	 */
	mp->reg[OffPitch] = (mode->x/8)<<22;

	mp->reg[OvrClr] = Pblack;
/* 8-bit DAC (won't work: kernel assumes 6-bit)
	if (mode->z >= 8)
		mp->reg[DacCntl] |= 0x100;
	else
		mp->reg[DacCntl] &= ~0x100;
*/
	if(ctlr->flag&Uenhanced)
		setdsp(vga, ctlr);

	if(vga->linear && mode->z != 1)
		ctlr->flag |= Ulinear;
	ctlr->flag |= Finit;
}

static void
load(Vga* vga, Ctlr* ctlr)
{
	Mach64xx *mp;

	mp = vga->private;

	/*
	 * Always use an 8Mb aperture on a 16Mb boundary.
	 */
	if(ctlr->flag & Ulinear)
		mp->reg[ConfigCntl] = ((vga->vmb/(4*1024*1024))<<4)|0x02;

	mp->iow32(mp, ConfigCntl, mp->reg[ConfigCntl]);

	mp->iow32(mp, GenTestCntl, 0);
	mp->iow32(mp, GenTestCntl, 0x100);

	if((ctlr->flag&Uenhanced) == 0)
	  mp->iow32(mp, MemCntl, mp->reg[MemCntl] & ~0x70000);
	mp->iow32(mp, HTotalDisp, mp->reg[HTotalDisp]);
	mp->iow32(mp, HSyncStrtWid, mp->reg[HSyncStrtWid]);
	mp->iow32(mp, VTotalDisp, mp->reg[VTotalDisp]);
	mp->iow32(mp, VSyncStrtWid, mp->reg[VSyncStrtWid]);
	mp->iow32(mp, IntCntl, mp->reg[IntCntl]);
	mp->iow32(mp, OffPitch, mp->reg[OffPitch]);
	mp->iow32(mp, CrtcGenCntl, mp->reg[CrtcGenCntl]);
	mp->iow32(mp, OvrClr, mp->reg[OvrClr]);
	if(ctlr->flag&Uenhanced){
	  mp->iow32(mp, DacCntl, mp->reg[DacCntl]);
	  mp->iow32(mp, CrtcGenCntl, mp->reg[CrtcGenCntl]&~0x02000000);
	  mp->iow32(mp, DspOnOff, mp->reg[DspOnOff]);
	  mp->iow32(mp, DspConfig, mp->reg[DspConfig]);
	  mp->iow32(mp, CrtcGenCntl, mp->reg[CrtcGenCntl]);
	  pllw(mp, PLLx, mp->pll[PLLx]);
	}
	pllw(mp, PLLn2, mp->pll[PLLn2]);
	pllw(mp, PLLp, mp->pll[PLLp]);
	mp->iow32(mp, ClockCntl, mp->reg[ClockCntl]);
	mp->iow32(mp, ClockCntl, 0x40|mp->reg[ClockCntl]);

	mp->iow32(mp, DpPixWidth, mp->reg[DpPixWidth]);

	if(vga->mode->z > 8){
		long sh, i;
		/*
		 * We need to initialize the palette, since the DACs use it
		 * in true color modes.  First see if the card supports 8-bit DAC.
		 */
		mp->iow32(mp, DacCntl, mp->reg[DacCntl] | 0x100);
		if(mp->ior32(mp, DacCntl)&0x100){
			/* card appears to support it */
			vgactlw("palettedepth", "8");
			mp->reg[DacCntl] |= 0x100;
		}

		if(mp->reg[DacCntl] & 0x100)
			sh = 0;	/* 8-bit DAC */
		else
			sh = 2;	/* 6-bit DAC */

		for(i=0; i<256; i++)
			setpalette(i, i>>sh, i>>sh, i>>sh);
	}

	ctlr->flag |= Fload;
}

static char* iorname[Nreg] = {
	"HTotalDisp",
	"HSyncStrtWid",
	"VTotalDisp",
	"VSyncStrtWid",
	"VlineCrntVline",
	"OffPitch",
	"IntCntl",
	"CrtcGenCntl",
	"LcdGenCntl",
	"PowerMngmnt",

	"OvrClr",
	"OvrWidLR",
	"OvrWidTB",

	"CurClr0",
	"CurClr1",
	"CurOffset",
	"CurHVposn",
	"CurHVoff",

	"ScratchReg0",
	"ScratchReg1",
	"ClockCntl",
	"BusCntl",
	"MemCntl",
	"MemVgaWpSel",
	"MemVgaRpSel",
	"DacRegs",
	"DacCntl",
	"GenTestCntl",
	"ConfigCntl",
	"ConfigChipId",
	"ConfigStat0",
	"ConfigStat1",
	"DspConfig",
	"DspOnOff",

	"DpBkgdClr",
	"DpChainMsk",
	"DpFrgdClr",
	"DpMix",
	"DpPixWidth",
	"DpSrc",
	"DpWriteMsk",	
};

static char* lcdname[Nlcd] = {
	"LCD_ConfigPanel",
	"LCD_GenCntl",
	"LCD_DstnCntl",
	"LCD_HfbPitchAddr",
	"LCD_HorzStretch",
	"LCD_VertStretch",
	"LCD_ExtVertStretch",
	"LCD_LtGio",
	"LCD_PowerMngmnt",
	"LCD_ZvgPio",
};

static void
dump(Vga* vga, Ctlr* ctlr)
{
	Mach64xx *mp;
	long i, m, n, p;
	double f;

	if((mp = vga->private) == 0)
		return;

	fprintf(stdout, "pci %p regtable %s io %lx %s\n", mp->pci, mp->regtable == pciregs ? "pciregs" : "ioregs", mp->io, mp->ior32 == mmior32 ? "mmio" : "portio");
	if(mp->pci)
		fprintf(stdout, "ccru %.4hx\n", mp->pci->ccru);
	for(i = 0; i < Nreg; i++)
		fprintf(stdout, "%s %-*s%.8lX\n", ctlr->name, 20, iorname[i], mp->reg[i]);

	printitem(ctlr->name, "PLL");
	for(i = 0; i < Npll; i++)
		printreg(mp->pll[i]);
	fprintf(stdout, "\n");

	for(i = 0; i < Nlcd; i++)
		fprintf(stdout, "%s %-*s%.8lX\n", ctlr->name, 30, lcdname[i], mp->lcd[i]);

	/*
	 *     (2*r*n)
	 * f = -------
	 *	(m*p)
	 */
	m = mp->pll[2];
	for(i = 0; i < 4; i++){
		n = mp->pll[7+i];
		p = 1<<((mp->pll[6]>>(i*2)) & 0x03);
		f = (2.0*RefFreq*n)/(m*p) + 0.5;
		fprintf(stdout, "%s VCLK%d\t%lu\n", ctlr->name, i, (long)f);
	}
}

static void
Cdump(Vga* vga, Ctlr* ctlr)
{
	Mach64xx *mp;
	long i, m, n, p;
	double f;

	if((mp = vga->private) == 0)
		return;

	for(i = 0; i < Nreg; i++)
		printf("mp->reg[%s] = 0x%lX;\n", iorname[i], mp->reg[i]);

	for(i = 0; i < Npll; i++)
		printf("mp->pll[0x%x] = 0x%x;\n", i, mp->pll[i]);

	for(i = 0; i < Nlcd; i++)
		printf("mp->lcd[%s] = 0x%lX;\n", lcdname[i], mp->lcd[i]);

	/*
	 *     (2*r*n)
	 * f = -------
	 *	(m*p)
	 */
	m = mp->pll[2];
	for(i = 0; i < 4; i++){
		n = mp->pll[7+i];
		p = 1<<((mp->pll[6]>>(i*2)) & 0x03);
		f = (2.0*RefFreq*n)/(m*p) + 0.5;
		fprintf(stdout, "// %s VCLK%d\t%lu\n", ctlr->name, i, (long)f);
	}
}

Ctlr mach64xx = {
	"mach64xx",			/* name */
	snarf,				/* snarf */
	0,				/* options */
	init,				/* init */
	load,				/* load */
	dump,				/* dump */
};

Ctlr mach64xxhwgc = {
	"mach64xxhwgc",			/* name */
	0,				/* snarf */
	0,				/* options */
	0,				/* init */
	0,				/* load */
	0,				/* dump */
};

