#include <u.h>
#include <libc.h>

#include "vga.h"


static char *usage =
	"usage: %s [-c clock] [-h hwgc] [-r ramdac] [-l link] controller\n";

static char *dbname = "/lib/vgadb";
static char monitordb[128];

static void
dump(Vga* vga)
{
	Ctlr *ctlr;

	if(vga->mode)
		dbdumpmode(vga->mode);

	for(ctlr = vga->link; ctlr; ctlr = ctlr->link){
		if(ctlr->dump == 0)
			continue;

		trace("%s->dump\n", ctlr->name);
		if(ctlr->flag && ctlr->flag != Fsnarf){
			printitem(ctlr->name, "flag");
			printflag(ctlr->flag);
			fprintf(stdout, "\n");
		}
		(*ctlr->dump)(vga, ctlr);
		ctlr->flag |= Fdump;
	}
	fprintf(stdout, "\n");
}

void
resyncinit(Vga* vga, Ctlr* ctlr, ulong on, ulong off)
{
	Ctlr *link;

	trace("%s->resyncinit on 0x%8.8lX off 0x%8.8lX\n", ctlr->name, on, off);

	for(link = vga->link; link; link = link->link){
		link->flag |= on;
		link->flag &= ~off;
		if(link == ctlr)
			continue;

		if(link->init == 0 || (link->flag & Finit) == 0)
			continue;
		link->flag &= ~Finit;
		trace("%s->init 0x%8.8lX\n", link->name, link->flag);
		(*link->init)(vga, link);
	}
}

void
sequencer(Vga* vga, long on)
{
	static uchar seq01;
	static long state = 1;
	char *s;

	if(on)
		s = "on";
	else
		s = "off";
	trace("sequencer->enter %s\n", s);
	if(on){
		if(vga)
			seq01 = vga->sequencer[0x01];
		if(state == 0){
			seq01 |= 0x01;
			vgaxo(Seqx, 0x01, seq01);
			vgaxo(Seqx, 0x00, 0x03);
		}
	}
	else{
		vgaxo(Seqx, 0x00, 0x01);
		seq01 = vgaxi(Seqx, 0x01);
		vgaxo(Seqx, 0x01, seq01|0x20);
	}
	state = on;
	trace("sequencer->leave %s\n", s);
}

char*
chanstr[32+1] = {
0,
/*1*/	"k1",
/*2*/	"k2",
0,
/*4*/	"k4",
0,
0,
0,
/*8*/	"m8",
0,
0,
0,
0,
0,
0,
0,
/*16*/	"r5g6b5",
0,
0,
0,
0,
0,
0,
0,
/*24*/	"r8g8b8",
0,
0,
0,
0,
0,
0,
0,
/*32*/	"a8r8g8b8",
};

char *ramdacstr;
char *clockstr;
char *linkstr;
char *hwgcstr;
char *argv0;

void
main(int argc, char** argv)
{
	char *vganame, buf[256], *p;
	long len;
	Ctlr *ctlr;
	Vga *vga;

//	Binit(stdout, 1, OWRITE);

	ARGBEGIN{
	case 'r':
		ramdacstr = ARGF();
		break;
	case 'c':
		clockstr = ARGF();
		break;
	case 'l':
		linkstr = ARGF();
		break;
	case 'h':
		hwgcstr = ARGF();
		break;
	case 'v':
		vflag++;
		break;
	default:
		error(usage, argv0);

	}ARGEND

	pcihinv(nil);

	if(argc != 1)
		error(usage, argv0);

	vganame = argv[0];

	vga = alloc(sizeof(Vga));

	dbctlr(vganame, vga);
	if(vga->ctlr == 0){
		fprintf(stdout, "%s: controller not found in lists\n", vganame);
		vga->ctlr = &generic;
		vga->link = &generic;
	}

	trace("main->snarf\n", 0);
	for(ctlr = vga->link; ctlr; ctlr = ctlr->link){
		if(ctlr->snarf == 0)
			continue;
		trace("%s->snarf\n", ctlr->name);
		(*ctlr->snarf)(vga, ctlr);
	}

	dump(vga);

	trace("main->exits\n", 0);
	exits(0);
}
